"""
The source code of this file was copied from the Keras project, and has been modified. All modifications
made from the original source code are under the LGPLv3 license.

COPYRIGHT

All contributions by François Chollet:
Copyright (c) 2015, François Chollet.
All rights reserved.

All contributions by Google:
Copyright (c) 2015, Google, Inc.
All rights reserved.

All contributions by Microsoft:
Copyright (c) 2017, Microsoft, Inc.
All rights reserved.

All other contributions:
Copyright (c) 2015 - 2017, the respective contributors.
All rights reserved.

Each contributor holds copyright over their respective contributions. The project versioning (Git)
records all such contribution source information.

LICENSE

The MIT License (MIT)

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES
OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from typing import Dict, List


class Callback:
    # pylint: disable=too-many-public-methods
    """
    Attributes:
        params (dict): Contains ``'epoch'`` and ``'steps_per_epoch'`` keys which are passed to the
            when training. Contains ``'steps'`` when evaluating. May contain other keys.
        model (Model): A reference to the :class:`~poutyne.Model` object which is using the callback.
    """

    def __init__(self):
        self.model = None
        self.params = None

    def set_params(self, params: Dict):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        """
        Is called before the beginning of each epoch.

        Args:
            epoch_number (int): The epoch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        """
        Is called before the end of each epoch.

        Args:
            epoch_number (int): The epoch number.
            logs (dict): Contains the following keys:

                 * ``'epoch'``: The epoch number.
                 * ``'time'``: The computation time of the epoch.
                 * ``'loss'``: The average loss of the batches.
                 * Values of training metrics: One key for each type of metrics. The metrics are also averaged.
                 * ``'val_loss'``: The average loss of the batches on the validation set.
                 * Values of validation metrics: One key for each type of metrics on the validation set. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'epoch': 2, 'time': 6.08248, 'loss': 0.40161, 'acc': 89.052, 'fscore_micro': 0.89051,
                    'val_loss': 0.36814, 'val_acc': 89.52, 'val_fscore_micro': 0.89520}
        """
        pass

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the training batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the training batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``'loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'loss': 1.95204, 'acc': 43.75}
        """
        pass

    def on_valid_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the validation batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the validation batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``val_loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'val_loss': 1.95204, 'val_acc': 43.75}
        """
        pass

    def on_test_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the testing batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_test_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the testing batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'size'``: The size of the batch as inferred by :func:`~Model.get_batch_size()`.
                 * ``'time'``: The computation time of the batch.
                 * ``'loss'``: The loss of the batch.
                 * Values of the batch metrics for the specific batch: One key for each type of metrics. Each key is
                   prefixed by ``'test_'``. The metrics are also averaged.

        Example::

            logs = {'batch': 171, 'size': 32, 'time': 0.00310, 'test_loss': 1.95204, 'test_acc': 43.75}
        """
        pass

    def on_predict_batch_begin(self, batch_number: int, logs: Dict):
        """
        Is called before the beginning of the predict batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Usually an empty dict.
        """
        pass

    def on_predict_batch_end(self, batch_number: int, logs: Dict):
        """
        Is called before the end of the predict batch.

        Args:
            batch_number (int): The batch number.
            logs (dict): Contains the following keys:

                 * ``'batch'``: The batch number.
                 * ``'time'``: The computation time of the batch.

        Example::

            logs = {'batch': 171, 'time': 0.00310}
        """
        pass

    def on_train_begin(self, logs: Dict):
        """
        Is called before the beginning of the training.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_train_end(self, logs: Dict):
        """
        Is called before the end of the training.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_begin(self, logs: Dict):
        """
        Is called before the beginning of the validation.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_valid_end(self, logs: Dict):
        """
        Is called before the end of the validation.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the test.
                 * ``'val_loss'``: The average loss of the batches on the test set.
                 * Values of testing metrics: One key for each type of metrics. Each key is
                   prefixed by ``'val_'``. The metrics are also averaged.

        Example::

            logs = {'time': 6.08248, 'val_loss': 0.40161, 'val_acc': 89.052, 'val_fscore_micro': 0.89051}
        """
        pass

    def on_test_begin(self, logs: Dict):
        """
        Is called before the beginning of the testing.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_test_end(self, logs: Dict):
        """
        Is called before the end of the testing.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the test.
                 * ``'test_loss'``: The average loss of the batches on the test set.
                 * Values of testing metrics: One key for each type of metrics. Each key is
                   prefixed by ``'test_'``. The metrics are also averaged.

        Example::

            logs = {'time': 6.08248, 'test_loss': 0.40161, 'test_acc': 89.052, 'test_fscore_micro': 0.89051}
        """
        pass

    def on_predict_begin(self, logs: Dict):
        """
        Is called before the beginning of the predict.

        Args:
            logs (dict): Usually an empty dict.
        """
        pass

    def on_predict_end(self, logs: Dict):
        """
        Is called before the end of the predict.

        Args:
            logs (dict): Contains the following keys:

                 * ``'time'``: The total computation time of the predict.

        Example::

            logs = {'time': 6.08248}
        """
        pass

    def on_backward_end(self, batch_number: int):
        """
        Is called after the backpropagation but before the optimization step.

        Args:
            batch_number (int): The batch number.
        """
        pass


class CallbackList:
    # pylint: disable=too-many-public-methods
    def __init__(self, callbacks: List[Callback]):
        callbacks = callbacks or []
        self.callbacks = list(callbacks)

    def append(self, callback: Callback):
        self.callbacks.append(callback)

    def set_params(self, params: Dict):
        for callback in self.callbacks:
            callback.set_params(params)

    def set_model(self, model):
        for callback in self.callbacks:
            callback.set_model(model)

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_begin(epoch_number, logs)

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_epoch_end(epoch_number, logs)

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_batch_begin(batch_number, logs)

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_batch_end(batch_number, logs)

    def on_valid_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_batch_begin(batch_number, logs)

    def on_valid_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_batch_end(batch_number, logs)

    def on_test_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_batch_begin(batch_number, logs)

    def on_test_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_batch_end(batch_number, logs)

    def on_predict_batch_begin(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_batch_begin(batch_number, logs)

    def on_predict_batch_end(self, batch_number: int, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_batch_end(batch_number, logs)

    def on_train_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_begin(logs)

    def on_train_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_train_end(logs)

    def on_valid_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_begin(logs)

    def on_valid_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_valid_end(logs)

    def on_test_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_begin(logs)

    def on_test_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_test_end(logs)

    def on_predict_begin(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_begin(logs)

    def on_predict_end(self, logs: Dict):
        logs = logs or {}
        for callback in self.callbacks:
            callback.on_predict_end(logs)

    def on_backward_end(self, batch_number: int):
        for callback in self.callbacks:
            callback.on_backward_end(batch_number)

    def __iter__(self):
        return iter(self.callbacks)



from .utils import camel_to_snake


def _get_registering_decorator(register_function):
    def register(name_or_func, *extra_names):
        if isinstance(name_or_func, str):
            names = [name_or_func] + list(extra_names)

            def decorator_func(func):
                register_function(func, names)
                return func

            return decorator_func

        func = name_or_func
        register_function(func)
        return func

    return register


batch_metrics_dict = {}


def clean_batch_metric_name(name):
    name = name.lower()
    name = name[:-4] if name.endswith('loss') else name
    name = name.replace('_', '')
    return name


def register_batch_metric_function(func, names=None):
    names = [func.__name__] if names is None else names
    names = [names] if isinstance(names, str) else names
    batch_metrics_dict.update({clean_batch_metric_name(name): func for name in names})


register_batch_metric = _get_registering_decorator(register_batch_metric_function)


def get_loss_or_metric(loss_metric):
    if isinstance(loss_metric, str):
        loss_metric = clean_batch_metric_name(loss_metric)
        return batch_metrics_dict[loss_metric]
    if isinstance(loss_metric, tuple) and isinstance(loss_metric[1], str):
        name, loss_metric = loss_metric
        loss_metric = clean_batch_metric_name(loss_metric)
        return name, batch_metrics_dict[loss_metric]
    return loss_metric


epochs_metrics_dict = {}


def clean_epoch_metric_name(name):
    name = name.lower()
    name = name[:-5] if name.endswith('score') else name
    name = name.replace('_', '')
    return name


def register_epoch_metric_class(clz, names=None):
    names = [camel_to_snake(clz.__name__)] if names is None else names
    names = [names] if isinstance(names, str) else names
    epochs_metrics_dict.update({clean_epoch_metric_name(name): clz for name in names})


register_epoch_metric = _get_registering_decorator(register_epoch_metric_class)


def get_epoch_metric(epoch_metric):
    if isinstance(epoch_metric, str):
        epoch_metric = clean_epoch_metric_name(epoch_metric)
        return epochs_metrics_dict[epoch_metric]()
    if isinstance(epoch_metric, tuple) and isinstance(epoch_metric[1], str):
        name, epoch_metric = epoch_metric
        epoch_metric = clean_epoch_metric_name(epoch_metric)
        return name, epochs_metrics_dict[epoch_metric]()
    return epoch_metric



import torch.optim as optim

all_optimizers_dict = dict(
    adadelta=optim.Adadelta,
    adagrad=optim.Adagrad,
    adam=optim.Adam,
    sparseadam=optim.SparseAdam,
    adamax=optim.Adamax,
    asgd=optim.ASGD,
    lbfgs=optim.LBFGS,
    rmsprop=optim.RMSprop,
    rprop=optim.Rprop,
    sgd=optim.SGD,
)


def get_optimizer(optimizer, module):
    if isinstance(optimizer, (str, dict)):
        kwargs = {}
        if isinstance(optimizer, dict):
            optimizer = dict(optimizer)
            kwargs = optimizer
            optimizer = optimizer.pop('optim')

        optimizer = optimizer.lower()

        if optimizer == 'sgd':
            kwargs.setdefault('lr', 1e-2)

        params = (p for p in module.parameters() if p.requires_grad)
        return all_optimizers_dict[optimizer](params, **kwargs)

    return optimizer



import math
from typing import Optional

BAR_EIGHTHS_CHARACTERS = [" ", "\u258f", "\u258e", "\u258d", "\u258c", "\u258b", "\u258a", "\u2589"]


class ProgressBar:
    """
    ProgressBar class keeps an update of the iteration and output the corresponding update progress bar.

    Args:
        steps (int): The number of steps.
        bar_format (str, optional): User defined format for the bar_format. By default the setting is
            {percentage} |{bar}|. The two argument must be {percentage} and {bar}.
        bar_character (str): The bar character to use. (Default value = \u2588 which is 'block'.)
        partial_bar_characters (str): A list of character to use when a fraction of the bar character
            could be used. (Default value =
            ``[" ", "\u258f", "\u258e", "\u258d", "\u258c", "\u258b", "\u258a", "\u2589"]``
            that is from " " through 1/8th to 7/8th blocks)

    Attributes:
        total_steps (int): The total steps to do.
        bar_format (str): The format of the bar.
        bar_character (str): The bar character.
        bar_length (int): The size of the bar.
        actual_steps (int): Number of steps done so far.
    """

    def __init__(
        self,
        steps: int,
        *,
        bar_length: int = 20,
        bar_format: Optional[str] = None,
        bar_character: str = "\u2588",
        partial_bar_characters: Optional[list] = None,
    ) -> None:
        self.total_steps = steps

        if bar_format is not None:
            self.bar_format = bar_format
        else:
            self.bar_format = "{percentage} |{bar}"

        self.bar_character = bar_character
        self.partial_bar_characters = BAR_EIGHTHS_CHARACTERS
        if partial_bar_characters is not None:
            self.partial_bar_characters = partial_bar_characters
        self.bar_length = bar_length

        self.actual_steps = 0

    def reset(self) -> None:
        """
        Reset the counter of the progress bar.
        """
        self.actual_steps = 0

    def update(self, n: int = 1) -> None:
        """
        To update the progress bar.

        Args:
            Value of the update. (Default value = 1)
        """
        self.actual_steps += n

    def __str__(self):
        """
        To format the progress bar for an output.
        """
        percentage = f"{self.actual_steps / self.total_steps * 100:6.2f}%"

        progress_bar = self.progress_bar_formatting()

        return self.bar_format.format(percentage=percentage, bar=progress_bar)

    def progress_bar_formatting(self) -> str:
        percentage = self.actual_steps / self.total_steps
        bar_length_complete = int(math.floor(percentage * self.bar_length)) * self.bar_character
        bar_length_incomplete = (self.bar_length - len(bar_length_complete) - 1) * " "

        partial_character = ""
        if percentage < 1.0:
            partial_percentage = percentage * self.bar_length - len(bar_length_complete)
            partial_index = int(math.floor(partial_percentage * len(self.partial_bar_characters)))
            partial_character = self.partial_bar_characters[partial_index]

        progress_bar = bar_length_complete + partial_character + bar_length_incomplete
        return progress_bar



import csv
import os
from typing import Dict, Optional, TextIO

from ._utils import atomic_lambda_save
from .callbacks import Callback


class Logger(Callback):
    def __init__(self, *, batch_granularity: bool = False):
        super().__init__()
        self.batch_granularity = batch_granularity
        self.epoch_number = 0

    def on_train_begin(self, logs: Dict):
        metrics = ['loss'] + self.model.metrics_names

        if self.batch_granularity:
            self.fieldnames = ['epoch', 'batch', 'size', 'time']
        else:
            self.fieldnames = ['epoch', 'time']
        if getattr(self.model, 'optimizer', None) is not None:
            if len(self.model.optimizer.param_groups) > 1:
                self.fieldnames += [f'lr_group_{i}' for i in range(len(self.model.optimizer.param_groups))]
            else:
                self.fieldnames += ['lr']
        self.fieldnames += metrics
        self.fieldnames += ['val_' + metric for metric in metrics]
        self._on_train_begin_write(logs)

    def _on_train_begin_write(self, logs: Dict):
        pass

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        if self.batch_granularity:
            logs = self._get_logs_without_unknown_keys(logs)
            self._on_train_batch_end_write(batch_number, logs)

    def _on_train_batch_end_write(self, batch_number: int, logs: Dict):
        pass

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        self.epoch_number = epoch_number
        self._on_epoch_begin_write(self.epoch_number, logs)

    def _on_epoch_begin_write(self, epoch_number: int, logs: Dict):
        pass

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        logs = self._get_logs_without_unknown_keys(logs)
        self._on_epoch_end_write(epoch_number, logs)

    def _on_epoch_end_write(self, epoch_number: int, logs: Dict):
        pass

    def on_train_end(self, logs: Dict):
        self._on_train_end_write(logs)

    def _on_train_end_write(self, logs: Dict):
        pass

    def _get_logs_without_unknown_keys(self, logs: Dict):
        return {k: logs[k] for k in self.fieldnames if logs.get(k) is not None}

    def _get_current_learning_rates(self):
        learning_rates = {}
        if getattr(self.model, 'optimizer', None) is not None:
            if len(self.model.optimizer.param_groups) > 1:
                learning_rates = {
                    f'lr_group_{i}': param_group['lr']
                    for i, param_group in enumerate(self.model.optimizer.param_groups)
                }
            else:
                learning_rates = {'lr': self.model.optimizer.param_groups[0]['lr']}
        return learning_rates


class CSVLogger(Logger):
    """
    Callback that outputs the result of each epoch_number or batch into a CSV file.

    Args:
        filename (str): The filename of the CSV.
        batch_granularity (bool): Whether to also output the result of each batch in addition to the epochs.
            (Default value = False)
        separator (str): The separator to use in the CSV.
            (Default value = ',')
        append (bool): Whether to append to an existing file.

    """

    def __init__(self, filename: str, *, batch_granularity: bool = False, separator: str = ',', append: bool = False):
        super().__init__(batch_granularity=batch_granularity)
        self.filename = filename
        self.separator = separator
        self.append = append

    def _on_train_begin_write(self, logs: Dict):
        open_flag = 'a' if self.append else 'w'
        self.csvfile = open(self.filename, open_flag, newline='', encoding='utf-8')
        self.writer = csv.DictWriter(self.csvfile, fieldnames=self.fieldnames, delimiter=self.separator)
        if not self.append:
            self.writer.writeheader()
            self.csvfile.flush()

    def _on_train_batch_end_write(self, batch_number: int, logs: Dict):
        self.writer.writerow(logs)
        self.csvfile.flush()

    def _on_epoch_end_write(self, epoch_number: int, logs: Dict):
        self.writer.writerow({**logs, **self._get_current_learning_rates()})
        self.csvfile.flush()

    def _on_train_end_write(self, logs: Dict):
        self.csvfile.close()


class AtomicCSVLogger(Logger):
    """
    Callback that outputs the result of each epoch_number or batch into a CSV file in an atomic matter.

    Args:
        filename (str): The filename of the CSV.
        temporary_filename (str, optional): Temporary filename for the CSV file so that it can be written
            atomically.
        batch_granularity (bool): Whether to also output the result of each batch in addition to the epochs.
            (Default value = False)
        separator (str): The separator to use in the CSV.
            (Default value = ',')
        append (bool): Whether to append to an existing file.
    """

    def __init__(
        self,
        filename,
        *,
        batch_granularity: bool = False,
        separator: str = ',',
        append: bool = False,
        temporary_filename: Optional[str] = None,
    ):
        super().__init__(batch_granularity=batch_granularity)
        self.filename = filename
        self.temporary_filename = temporary_filename
        self.separator = separator
        self.append = append

    def _save_log(self, fd: TextIO, logs: Dict):
        olddata = None
        if os.path.exists(self.filename):
            with open(self.filename, 'r', encoding='utf-8') as oldfile:
                olddata = list(csv.DictReader(oldfile, delimiter=self.separator))
        csvwriter = csv.DictWriter(fd, fieldnames=self.fieldnames, delimiter=self.separator)
        csvwriter.writeheader()
        if olddata is not None:
            csvwriter.writerows(olddata)
        if logs is not None:
            csvwriter.writerow(logs)

    def _write_header(self, fd: TextIO):
        csvwriter = csv.DictWriter(fd, fieldnames=self.fieldnames, delimiter=self.separator)
        csvwriter.writeheader()

    def _on_train_begin_write(self, logs: Dict):
        if not self.append:
            atomic_lambda_save(self.filename, self._write_header, (), temporary_filename=self.temporary_filename)

    def _on_train_batch_end_write(self, batch_number: int, logs: Dict):
        atomic_lambda_save(self.filename, self._save_log, (logs,), temporary_filename=self.temporary_filename)

    def _on_epoch_end_write(self, epoch_number: int, logs: Dict):
        logs = {**logs, **self._get_current_learning_rates()}
        atomic_lambda_save(self.filename, self._save_log, (logs,), temporary_filename=self.temporary_filename)


class TensorBoardLogger(Logger):
    """
    Callback that outputs the result of each epoch_number or batch into a Tensorboard experiment folder.

    Args:
        writer (~torch.utils.tensorboard.writer.SummaryWriter): The tensorboard writer.

    Example:
        Using TensorBoardLogger::

            from torch.utils.tensorboard import SummaryWriter
            from poutyne import Model, TensorBoardLogger

            writer = SummaryWriter('runs')
            tb_logger = TensorBoardLogger(writer)

            model = Model(...)
            model.fit_generator(..., callbacks=[tb_logger])
    """

    def __init__(self, writer):
        super().__init__(batch_granularity=False)
        self.writer = writer

    def _on_train_batch_end_write(self, batch_number: int, logs):
        """
        We don't handle tensorboard writing on batch granularity
        """
        pass

    def _on_epoch_end_write(self, epoch_number: int, logs: dict):
        grouped_items = {}
        for k, v in logs.items():
            if 'val_' in k:
                primary_key = k[4:]
                if primary_key not in grouped_items:
                    grouped_items[primary_key] = {}
                grouped_items[k[4:]][k] = v
            else:
                if k not in grouped_items:
                    grouped_items[k] = {}
                grouped_items[k][k] = v
        for k, v in grouped_items.items():
            self.writer.add_scalars(k, v, epoch_number)
        lr = self._get_current_learning_rates()
        self.writer.add_scalars('lr', lr, epoch_number)

    def _on_train_end_write(self, logs: Dict):
        self.writer.close()



import math
from tempfile import TemporaryDirectory
from unittest import TestCase, skipIf
from unittest.mock import MagicMock, call, ANY

import torch
import torch.nn as nn

from tests.framework.tools import some_data_generator

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    SummaryWriter = None

from poutyne import Model, TensorBoardGradientTracker, WeightsGradientsStatsTracker, Tracker


class TrackerTest(TestCase):
    def test_keep_good_layer(self):
        # pylint: disable=protected-access
        tracker = Tracker(keep_bias=False)
        layer_to_keep_params = MagicMock()
        layer_to_keep_params.requires_grad = True
        self.assertTrue(tracker._keep_layer(layer_to_keep_params, "fake_layer_name_to_keep"))
        self.assertFalse(tracker._keep_layer(layer_to_keep_params, "bias_name_not_to_keep"))

        layer_not_to_keep_params = MagicMock()
        layer_not_to_keep_params.requires_grad = False
        self.assertFalse(tracker._keep_layer(layer_not_to_keep_params, "fake_layer_name_not_to_keep"))
        self.assertFalse(tracker._keep_layer(layer_not_to_keep_params, "bias_name_not_to_keep"))

        tracker = Tracker(keep_bias=True)
        layer_to_keep_params = MagicMock()
        layer_to_keep_params.requires_grad = True
        self.assertTrue(tracker._keep_layer(layer_to_keep_params, "fake_layer_name_to_keep"))
        self.assertTrue(tracker._keep_layer(layer_to_keep_params, "bias_name_to_keep"))

        layer_not_to_keep_params = MagicMock()
        layer_not_to_keep_params.requires_grad = False
        self.assertFalse(tracker._keep_layer(layer_not_to_keep_params, "fake_layer_name_not_to_keep"))
        self.assertFalse(tracker._keep_layer(layer_not_to_keep_params, "bias_name_not_to_keep"))


class GradientStatsTrackerTest(TestCase):
    def setUp(self):
        self.tracker = WeightsGradientsStatsTracker(number_layers=2)

        self.absolute_min_both_layer = 0.00
        self.layer_1_min = -0.15
        self.layer_1_max = 0.24
        self.layer_1_gradients = torch.Tensor([[self.layer_1_max], [self.absolute_min_both_layer], [self.layer_1_min]])

        self.layer_2_min = -0.25
        self.layer_2_max = 0.16
        self.layer_2_gradients = torch.Tensor([[self.layer_2_max], [self.layer_2_min], [self.absolute_min_both_layer]])

        self.layer_1_name = "fake_name_1"
        self.layer_2_name = "fake_name_2"
        self.layer_names = [self.layer_1_name, self.layer_2_name]

        # The value have been compute manual according to the Welford's online algorithm
        # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
        # We refer to the WelfordCompute document in tests/framework/callbacks for the expected value computation.
        mean_1 = 0.13
        mean_2 = 0.1366666
        s2_1 = 0
        s2_2 = 0
        self.first_batch_expected_stats = {
            self.layer_1_name: {
                "mean": mean_1,
                "mean_std_dev_up": mean_1 + math.sqrt(s2_1),
                "mean_std_dev_down": mean_1 - math.sqrt(s2_1),
                "min": self.layer_1_min,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_1_max,
                "abs_max": self.layer_1_max,
            },
            self.layer_2_name: {
                "mean": mean_2,
                "mean_std_dev_up": mean_2 + math.sqrt(s2_2),
                "mean_std_dev_down": mean_2 - math.sqrt(s2_2),
                "min": self.layer_2_min,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_2_max,
                "abs_max": math.fabs(self.layer_2_min),  # since the absolute value is higher than the normal max
            },
        }

        mean_1 = 0.195
        mean_2 = 0.205
        s2_1 = 0.00845
        s2_2 = 0.0093889
        # *2 for min and max since the weights are 3 times the layers gradients
        self.second_batch_expected_stats = {
            self.layer_1_name: {
                "mean": mean_1,
                "mean_std_dev_up": mean_1 + math.sqrt(s2_1),
                "mean_std_dev_down": mean_1 - math.sqrt(s2_1),
                "min": self.layer_1_min * 2,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_1_max * 2,
                "abs_max": self.layer_1_max * 2,
            },
            self.layer_2_name: {
                "mean": mean_2,
                "mean_std_dev_up": mean_2 + math.sqrt(s2_2),
                "mean_std_dev_down": mean_2 - math.sqrt(s2_2),
                "min": self.layer_2_min * 2,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_2_max * 2,
                "abs_max": math.fabs(self.layer_2_min * 2),  # since the absolute value is higher than the normal max
            },
        }

        mean_1 = 0.26
        mean_2 = 0.27333333333
        s2_1 = 0.0169
        s2_2 = 0.018677778
        # *3 for min and max since the weights are 3 times the layers gradients
        self.third_batch_expected_stats = {
            self.layer_1_name: {
                "mean": mean_1,
                "mean_std_dev_up": mean_1 + math.sqrt(s2_1),
                "mean_std_dev_down": mean_1 - math.sqrt(s2_1),
                "min": self.layer_1_min * 3,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_1_max * 3,
                "abs_max": self.layer_1_max * 3,
            },
            self.layer_2_name: {
                "mean": mean_2,
                "mean_std_dev_up": mean_2 + math.sqrt(s2_2),
                "mean_std_dev_down": mean_2 - math.sqrt(s2_2),
                "min": self.layer_2_min * 3,
                "abs_min": self.absolute_min_both_layer,
                "max": self.layer_2_max * 3,
                "abs_max": math.fabs(self.layer_2_min * 3),  # since the absolute value is higher than the normal max
            },
        }

    def test_compute_two_layers_statistic(self):
        self._run_n_batch(num_batch=1)
        self._test_batch_stats(self.first_batch_expected_stats)

        self._run_n_batch(num_batch=2)
        self._test_batch_stats(self.second_batch_expected_stats)

        self._run_n_batch(num_batch=3)
        self._test_batch_stats(self.third_batch_expected_stats)

    def _run_n_batch(self, num_batch):
        for batch_number in range(1, num_batch + 1):
            layer_1_params = MagicMock()
            layer_1_params.grad = self.layer_1_gradients * batch_number
            layer_2_params = MagicMock()
            layer_2_params.grad = self.layer_2_gradients * batch_number

            named_parameters = (
                (n, p) for n, p in [(self.layer_1_name, layer_1_params), (self.layer_2_name, layer_2_params)]
            )

            self.tracker.batch_statistic_upgrade(named_parameters=named_parameters)

    def _test_batch_stats(self, batch_expected):
        batch_actual_stats = self.tracker.get_stats(self.layer_names)

        self._test_stats(batch_expected[self.layer_1_name], batch_actual_stats[self.layer_1_name])

        self._test_stats(batch_expected[self.layer_2_name], batch_actual_stats[self.layer_2_name])

    def _test_stats(self, expected, actual):
        self.assertEqual(len(expected), len(actual))
        self.assertEqual(expected.keys(), actual.keys())
        for expected_value, actual_value in zip(expected.values(), actual.values()):
            self.assertAlmostEqual(float(expected_value), float(actual_value), places=3)


@skipIf(SummaryWriter is None, "Unable to import SummaryWriter from torch")
class TensorBoardGradientTrackerTest(TestCase):
    batch_size = 20
    lr = 1e-3
    num_epochs = 10

    def setUp(self):
        torch.manual_seed(42)
        self.loss_function = nn.MSELoss()
        self.temp_dir_obj = TemporaryDirectory()
        # pylint: disable=not-callable
        self.writer = SummaryWriter(self.temp_dir_obj.name)
        self.writer.add_scalars = MagicMock()

    def tearDown(self):
        self.temp_dir_obj.cleanup()

    def test_tracking_one_layer_model(self):
        self.num_layer = 1
        self.pytorch_network = nn.Linear(1, 1)
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=self.lr)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

        keep_bias = False
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        tracker = TensorBoardGradientTracker(self.writer, keep_bias=keep_bias)
        self.model.fit_generator(train_gen, valid_gen, epochs=self.num_epochs, steps_per_epoch=5, callbacks=[tracker])
        self._test_tracking(keep_bias)

    def test_tracking_one_layer_model_with_bias(self):
        self.num_layer = 1
        self.pytorch_network = nn.Linear(1, 1)
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=self.lr)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

        keep_bias = True
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        tracker = TensorBoardGradientTracker(self.writer, keep_bias=keep_bias)
        self.model.fit_generator(train_gen, valid_gen, epochs=self.num_epochs, steps_per_epoch=5, callbacks=[tracker])
        self._test_tracking(keep_bias)

    def test_tracking_two_layers_model(self):
        self.num_layer = 2
        self.pytorch_network = nn.Sequential(nn.Linear(1, 1), nn.Linear(1, 1))
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=self.lr)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

        keep_bias = False
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        tracker = TensorBoardGradientTracker(self.writer, keep_bias=keep_bias)
        self.model.fit_generator(train_gen, valid_gen, epochs=self.num_epochs, steps_per_epoch=5, callbacks=[tracker])
        self._test_tracking(keep_bias)

    def test_tracking_two_layers_shallow_model(self):
        self.num_layer = 2
        self.pytorch_network = nn.Sequential(nn.Linear(1, 4), nn.Linear(4, 1))
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=self.lr)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

        keep_bias = False
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        tracker = TensorBoardGradientTracker(self.writer, keep_bias=keep_bias)
        self.model.fit_generator(train_gen, valid_gen, epochs=self.num_epochs, steps_per_epoch=5, callbacks=[tracker])
        self._test_tracking(keep_bias)

    def test_tracking_N_layers_model_with_bias(self):
        self.num_layer = 4
        self.pytorch_network = nn.Sequential(nn.Linear(1, 1), nn.Linear(1, 1), nn.Linear(1, 1), nn.Linear(1, 1))
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=self.lr)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

        keep_bias = True
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        tracker = TensorBoardGradientTracker(self.writer, keep_bias=keep_bias)
        self.model.fit_generator(train_gen, valid_gen, epochs=self.num_epochs, steps_per_epoch=5, callbacks=[tracker])
        self._test_tracking(keep_bias)

    def _test_tracking(self, keep_bias):
        expected_calls = []
        for epoch in range(1, self.num_epochs + 1):
            layer_names = [""]
            if self.num_layer > 1:
                layer_names = []
                for layer_idx in range(self.num_layer):
                    layer_names.append(f"{layer_idx}.")
            for layer_name in layer_names:
                expected_calls.append(call(f'gradient_distributions/{layer_name}weight', {'mean': ANY}, epoch))
                expected_calls.append(
                    call(f'gradient_distributions/{layer_name}weight', {'mean_std_dev_up': ANY}, epoch)
                )
                expected_calls.append(
                    call(f'gradient_distributions/{layer_name}weight', {'mean_std_dev_down': ANY}, epoch)
                )
                expected_calls.append(call(f'other_gradient_stats/{layer_name}weight', {'min': ANY}, epoch))
                expected_calls.append(call(f'other_gradient_stats/{layer_name}weight', {'max': ANY}, epoch))

                if keep_bias:
                    expected_calls.append(call(f'gradient_distributions/{layer_name}bias', {'mean': ANY}, epoch))
                    expected_calls.append(
                        call(f'gradient_distributions/{layer_name}bias', {'mean_std_dev_up': ANY}, epoch)
                    )
                    expected_calls.append(
                        call(f'gradient_distributions/{layer_name}bias', {'mean_std_dev_down': ANY}, epoch)
                    )
                    expected_calls.append(call(f'other_gradient_stats/{layer_name}bias', {'min': ANY}, epoch))
                    expected_calls.append(call(f'other_gradient_stats/{layer_name}bias', {'max': ANY}, epoch))

        method_calls = self.writer.add_scalars.mock_calls
        self.assertEqual(len(method_calls), len(expected_calls))
        self.assertEqual(method_calls, expected_calls)

        self.assertIn(expected_calls, method_calls)



# -*- coding: utf-8 -*-
from unittest import TestCase
from unittest.mock import MagicMock, call
import unittest
import torch
import numpy as np
from poutyne.utils import _concat
from poutyne import TensorDataset, torch_apply


class TorchApplyTest(TestCase):
    def test_apply_on_list(self):
        my_list = [MagicMock(spec=torch.Tensor) for _ in range(10)]
        torch_apply(my_list, lambda t: t.cpu())
        self._test_method_calls(my_list)

    def test_apply_on_recursive_list(self):
        my_list = [MagicMock(spec=torch.Tensor) for _ in range(2)]
        my_list.append([MagicMock(spec=torch.Tensor) for _ in range(3)])
        my_list += [MagicMock(spec=torch.Tensor) for _ in range(1)]
        torch_apply(my_list, lambda t: t.cpu())
        self._test_method_calls(my_list[:2] + my_list[2] + my_list[3:])

    def test_apply_on_tuple(self):
        my_tuple = tuple(MagicMock(spec=torch.Tensor) for _ in range(10))
        torch_apply(my_tuple, lambda t: t.cpu())
        self._test_method_calls(my_tuple)

    def test_apply_on_recursive_tuple(self):
        my_tuple = tuple(MagicMock(spec=torch.Tensor) for _ in range(2))
        my_tuple += (tuple(MagicMock(spec=torch.Tensor) for _ in range(3)),)
        my_tuple += tuple(MagicMock(spec=torch.Tensor) for _ in range(1))
        torch_apply(my_tuple, lambda t: t.cpu())
        self._test_method_calls(my_tuple[:2] + my_tuple[2] + my_tuple[3:])

    def test_apply_on_dict(self):
        my_dict = {}
        for k in ['a', 'b', 'c']:
            my_dict[k] = MagicMock(spec=torch.Tensor)
        torch_apply(my_dict, lambda t: t.cpu())
        self._test_method_calls(list(my_dict.values()))

    def test_apply_on_recursive_dict(self):
        my_dict = {}
        my_dict['a'] = MagicMock(spec=torch.Tensor)
        my_dict['b'] = {}
        for k in ['c', 'd']:
            my_dict['b'][k] = MagicMock(spec=torch.Tensor)
        torch_apply(my_dict, lambda t: t.cpu())
        self._test_method_calls([my_dict['a'], *my_dict['b'].values()])

    def test_apply_on_recursive_data_structure(self):
        my_obj = {
            'a': [MagicMock(spec=torch.Tensor) for _ in range(3)],
            'b': tuple(MagicMock(spec=torch.Tensor) for _ in range(2)),
            'c': {'d': [MagicMock(spec=torch.Tensor) for _ in range(3)]},
            'e': MagicMock(spec=torch.Tensor),
        }
        torch_apply(my_obj, lambda t: t.cpu())
        self._test_method_calls(my_obj['a'] + list(my_obj['b']) + my_obj['c']['d'] + [my_obj['e']])

    def test_apply_on_object_with_no_tensor(self):
        my_obj = {'a': 5, 'b': 3.141592, 'c': {'d': [1, 2, 3]}}
        ret = torch_apply(my_obj, lambda t: t.cpu())
        self.assertEqual(ret, my_obj)
        self.assertFalse(ret is my_obj)

    def test_apply_with_replacement_to_no_tensor(self):
        my_obj = [MagicMock(spec=torch.Tensor)]
        ret = torch_apply(my_obj, lambda t: 123)
        self.assertEqual(ret, [123])

    def _test_method_calls(self, mock_list):
        print(mock_list)
        for mock in mock_list:
            self.assertEqual(mock.method_calls, [call.cpu()])


class TensorDatasetTest(TestCase):
    def test_one_tensor(self):
        range20 = np.expand_dims(np.arange(20), 1)
        dataset = TensorDataset(range20)
        self.assertEqual(len(dataset), 20)
        for i in range(20):
            self.assertEqual(dataset[i], np.array([i]))

    def test_multiple_tensors(self):
        range20 = np.expand_dims(np.arange(20), 1)
        dataset = TensorDataset(range20, range20 * 2, range20 * 3)
        self.assertEqual(len(dataset), 20)
        self.assertEqual(type(dataset[0]), tuple)
        for i in range(20):
            self.assertEqual(dataset[i][0], i)
            self.assertEqual(dataset[i][1], i * 2)
            self.assertEqual(dataset[i][2], i * 3)

    def test_list_of_tensors(self):
        range20 = np.expand_dims(np.arange(20), 1)
        dataset = TensorDataset((range20, range20 * 2), range20 * 3)
        self.assertEqual(len(dataset), 20)
        self.assertEqual(type(dataset[0]), tuple)
        self.assertEqual(type(dataset[0][0]), tuple)
        self.assertEqual(type(dataset[0][-1]), np.ndarray)
        for i in range(20):
            self.assertEqual(dataset[i][0][0], i)
            self.assertEqual(dataset[i][0][1], i * 2)
            self.assertEqual(dataset[i][1], i * 3)

        dataset = TensorDataset((range20, range20 * 2), (range20 * 3, range20 * 4))
        self.assertEqual(len(dataset), 20)

        self.assertEqual(type(dataset[0]), tuple)
        self.assertEqual(type(dataset[1]), tuple)
        self.assertEqual(type(dataset[0][0]), tuple)
        self.assertEqual(type(dataset[0][1]), tuple)
        for i in range(20):
            self.assertEqual(type(dataset[i][0][0]), np.ndarray)
            self.assertEqual(type(dataset[i][0][1]), np.ndarray)
            self.assertEqual(dataset[i][0][0], i)
            self.assertEqual(dataset[i][0][1], i * 2)
            self.assertEqual(type(dataset[i][1][0]), np.ndarray)
            self.assertEqual(type(dataset[i][1][1]), np.ndarray)
            self.assertEqual(dataset[i][1][0], i * 3)
            self.assertEqual(dataset[i][1][1], i * 4)


class ConcatTest(TestCase):
    def test_single_array(self):
        """
        Test the concatenation of a single array
        """
        obj = [np.arange(5)] * 5
        concat = _concat(obj)
        self.assertEqual(concat.shape, (25,))

    def test_tuple_1(self):
        """
        Test the concatenation of a [([], [])]
        """
        obj = [(np.arange(5), np.ones(5) * 2)] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0].shape, (25,))
        self.assertEqual(concat[1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][i * 5 + j] == j)
        self.assertTrue((concat[1] == 2).all())

    def test_tuple_2(self):
        """
        Test the concatenation of a [([], ([], []))]
        """
        obj = [(np.arange(5), (np.ones(5) * 2, np.ones(5) * 3))] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0].shape, (25,))
        self.assertEqual(concat[1][0].shape, (25,))
        self.assertEqual(concat[1][1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][i * 5 + j] == j)
        self.assertTrue((concat[1][0] == 2).all())
        self.assertTrue((concat[1][1] == 3).all())

    def test_tuple_3(self):
        """
        Test the concatenation of a [(([], []), ([], []))]
        """
        obj = [((np.arange(5), np.ones(5)), (np.ones(5) * 2, np.ones(5) * 3))] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0][0].shape, (25,))
        self.assertEqual(concat[0][1].shape, (25,))
        self.assertEqual(concat[1][0].shape, (25,))
        self.assertEqual(concat[1][1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][0][i * 5 + j] == j)
        self.assertTrue((concat[0][1] == 1).all())
        self.assertTrue((concat[1][0] == 2).all())
        self.assertTrue((concat[1][1] == 3).all())

    def test_array_1(self):
        """
        Test the concatenation of a [[[], []]]
        """
        obj = [[np.arange(5), np.ones(5) * 2]] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0].shape, (25,))
        self.assertEqual(concat[1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][i * 5 + j] == j)
        self.assertTrue((concat[1] == 2).all())

    def test_array_2(self):
        """
        Test the concatenation of a [[[], ([], [])]]
        """
        obj = [[np.arange(5), [np.ones(5) * 2, np.ones(5) * 3]]] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0].shape, (25,))
        self.assertEqual(concat[1][0].shape, (25,))
        self.assertEqual(concat[1][1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][i * 5 + j] == j)
        self.assertTrue((concat[1][0] == 2).all())
        self.assertTrue((concat[1][1] == 3).all())

    def test_array_3(self):
        """
        Test the concatenation of a [[[[], []], [[], []]]]
        """
        obj = [[[np.arange(5), np.ones(5)], [np.ones(5) * 2, np.ones(5) * 3]]] * 5
        concat = _concat(obj)
        self.assertEqual(concat[0][0].shape, (25,))
        self.assertEqual(concat[0][1].shape, (25,))
        self.assertEqual(concat[1][0].shape, (25,))
        self.assertEqual(concat[1][1].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat[0][0][i * 5 + j] == j)
        self.assertTrue((concat[0][1] == 1).all())
        self.assertTrue((concat[1][0] == 2).all())
        self.assertTrue((concat[1][1] == 3).all())

    def test_dict_1(self):
        """
        Test list of dictionaries
        """
        obj = [{'a': np.arange(5), 'b': np.ones(5) * 2}] * 5
        concat = _concat(obj)
        self.assertEqual(concat['a'].shape, (25,))
        self.assertEqual(concat['b'].shape, (25,))
        for i in range(5):
            for j in range(5):
                self.assertTrue(concat['a'][i * 5 + j] == j)
        self.assertTrue((concat['b'] == 2).all())

    def test_dict_2(self):
        """
        Test list of dictionaries
        """
        obj = [{'a': (np.arange(5), np.ones(5)), 'b': np.ones(5) * 2}] * 5
        concat = _concat(obj)
        self.assertEqual(concat['a'][0].shape, (25,))
        self.assertEqual(concat['a'][1].shape, (25,))
        self.assertEqual(concat['b'].shape, (25,))

        for i in range(5):
            for j in range(5):
                self.assertTrue(concat['a'][0][i * 5 + j] == j)
        self.assertTrue((concat['a'][1] == 1).all())
        self.assertTrue((concat['b'] == 2).all())

    def test_non_concatenable_values(self):
        obj = [3] * 5
        concat = _concat(obj)
        self.assertEqual(concat, obj)
        self.assertEqual(type(concat), type(obj))

    def test_non_concatenable_values2(self):
        obj = [{'a': (np.arange(5), np.ones(5), 2), 'b': 3, 'c': np.array(4)}] * 5
        concat = _concat(obj)
        self.assertEqual(concat['a'][0].shape, (25,))
        self.assertEqual(concat['a'][1].shape, (25,))
        self.assertEqual(concat['a'][2], (2,) * 5)
        self.assertEqual(concat['b'], [3] * 5)
        self.assertEqual(concat['c'], [4] * 5)

        for i in range(5):
            for j in range(5):
                self.assertTrue(concat['a'][0][i * 5 + j] == j)
        self.assertTrue((concat['a'][1] == 1).all())


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from typing import Dict, Union

from . import Callback


class Notificator(ABC):
    """
    The interface of the Notificator. It must at least implement a `send_notification` method.
    The interface is similar to the `notif <https://notificationdoc.ca/index.html>`_ package.
    """

    @abstractmethod
    def send_notification(self, message: str, *, subject: Union[str, None] = None) -> None:
        """
        Abstract method to send a notification.

        Args:
            message (str): The message to send as a notification message through the notificator.
            subject (str): The subject of the notification. If None, the default message is used. By default, None.
                Also, we recommend formatting the subject for better readability, e.g. using bolding it using Markdown
                and appending with a new line.
        """


class NotificationCallback(Callback):
    """
    Send a notification to a channel at the beginning/ending of the training/testing and at a constant frequency
    (`alert_frequency`) during the training.

    Args:
        notificator (~poutyne.Notificator): The notification channel to send the message.
            The expected interface need to implement a `send_notification` method to send the message. You can see the
            `notif <https://notificationdoc.ca/index.html>`_ package which implements some Notificator respecting the
            interface.
        alert_frequency (int): The frequency (in epoch), during training, to send an update. By default, 1.
        experiment_name (Union[str, None]): The name of the experiment to add to the message. By default, None.

    Example:

        .. code-block:: python

            from notif.notificator import SlackNotificator
            from poutyne.framework.callbacks.notification import NotificationCallback

            webhook_url = "a_link"
            slack_notif = SlackNotificator(webhook_url=webhook_url)

            notif_callback = NotificationCallback(notificator=slack_notif)

            model = Model(...)
            model.fit_generator(..., callbacks=[notif_callback])
    """

    def __init__(
        self, notificator: Notificator, alert_frequency: int = 1, experiment_name: Union[None, str] = None
    ) -> None:
        super().__init__()
        self.notificator = notificator
        self.alert_frequency = alert_frequency
        self.experiment_name_msg = f" for {experiment_name}" if experiment_name is not None else ""

    def on_train_begin(self, logs: Dict) -> None:
        """
        Send the message to the channel 'Start of the training' or
        'Start of the training for the experiment experiment_name' if an experiment name is given.
        """
        empty_message = ""
        self.notificator.send_notification(empty_message, subject=f"Start of the training{self.experiment_name_msg}.")

    def on_epoch_end(self, epoch_number: int, logs: Dict) -> None:
        """
        Send the message to the channel 'Epoch is done' or 'Epoch is done for the experiment experiment_name'
        if an experiment name is given and the logs metrics (one per line).
        """

        if epoch_number % self.alert_frequency == 0:
            message = f"Here the epoch metrics: \n{self._format_logs(logs)}"
            self.notificator.send_notification(
                message, subject=f"Epoch {epoch_number} is done{self.experiment_name_msg}."
            )

    def on_train_end(self, logs: Dict) -> None:
        """
        Send the message to the channel 'End of the training' or
        'End of the training for the experiment experiment_name' if an experiment name is given.
        """

        empty_message = ""
        self.notificator.send_notification(empty_message, subject=f"End of the training{self.experiment_name_msg}.")

    def on_test_begin(self, logs: Dict) -> None:
        """
        Send the message to the channel 'Start of the testing' or
        'Start of the testing for the experiment experiment_name' if an experiment name is given.
        """

        empty_message = ""
        self.notificator.send_notification(empty_message, subject=f"Start of the testing{self.experiment_name_msg}.")

    def on_test_end(self, logs: Dict) -> None:
        """
        Send the message to the channel 'End of the testing' or
        'End of the testing for the experiment experiment_name' if an experiment name is given.
        """

        message = f"Here the epoch metrics: \n{self._format_logs(logs)}"
        self.notificator.send_notification(message, subject=f"End of the testing{self.experiment_name_msg}.")

    @staticmethod
    def _format_logs(logs: Dict) -> str:
        return " ".join([f"{key}: {value}\n" for key, value in logs.items()])



import numpy as np
import torch
import torch.nn as nn

from poutyne import Model, warning_settings

from .base import ModelFittingTestCase, MultiIOModel

warning_settings['concatenate_returns'] = 'ignore'


def some_data_tensor_generator_multi_io(batch_size):
    while True:
        x1 = torch.rand(batch_size, 1)
        x2 = torch.rand(batch_size, 1)
        y1 = torch.rand(batch_size, 1)
        y2 = torch.rand(batch_size, 1)
        yield (x1, x2), (y1, y2)


class ModelMultiIOTest(ModelFittingTestCase):
    def setUp(self):
        super().setUp()
        torch.manual_seed(42)
        self.pytorch_network = MultiIOModel(num_input=2, num_output=2)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=1e-3)

        self.model = Model(
            self.pytorch_network,
            self.optimizer,
            lambda y_pred, y_true: self.loss_function(y_pred[0], y_true[0]) + self.loss_function(y_pred[1], y_true[1]),
            batch_metrics=self.batch_metrics,
            epoch_metrics=self.epoch_metrics,
        )

    def test_fitting_tensor_generator_multi_io(self):
        train_generator = some_data_tensor_generator_multi_io(ModelMultiIOTest.batch_size)
        valid_generator = some_data_tensor_generator_multi_io(ModelMultiIOTest.batch_size)
        logs = self.model.fit_generator(
            train_generator,
            valid_generator,
            epochs=ModelMultiIOTest.epochs,
            steps_per_epoch=ModelMultiIOTest.steps_per_epoch,
            validation_steps=ModelMultiIOTest.steps_per_epoch,
            callbacks=[self.mock_callback],
        )
        params = {
            'epochs': ModelMultiIOTest.epochs,
            'steps': ModelMultiIOTest.steps_per_epoch,
            'valid_steps': ModelMultiIOTest.steps_per_epoch,
        }
        self._test_callbacks_train(params, logs, valid_steps=ModelMultiIOTest.steps_per_epoch)

    def test_fitting_with_tensor_multi_io(self):
        train_real_steps_per_epoch = 30
        train_batch_size = ModelMultiIOTest.batch_size
        train_final_batch_missing_samples = 7
        train_size = train_real_steps_per_epoch * train_batch_size - train_final_batch_missing_samples
        train_x = (torch.rand(train_size, 1), torch.rand(train_size, 1))
        train_y = (torch.rand(train_size, 1), torch.rand(train_size, 1))

        valid_real_steps_per_epoch = 10
        # valid_batch_size will be the same as train_batch_size in the fit method.
        valid_batch_size = train_batch_size
        valid_final_batch_missing_samples = 3
        valid_size = valid_real_steps_per_epoch * valid_batch_size - valid_final_batch_missing_samples
        valid_x = (torch.rand(valid_size, 1), torch.rand(valid_size, 1))
        valid_y = (torch.rand(valid_size, 1), torch.rand(valid_size, 1))

        logs = self.model.fit(
            train_x,
            train_y,
            validation_data=(valid_x, valid_y),
            epochs=ModelMultiIOTest.epochs,
            batch_size=train_batch_size,
            steps_per_epoch=None,
            validation_steps=None,
            callbacks=[self.mock_callback],
        )
        params = {
            'epochs': ModelMultiIOTest.epochs,
            'steps': train_real_steps_per_epoch,
            'valid_steps': valid_real_steps_per_epoch,
        }
        self._test_callbacks_train(params, logs)

    def test_tensor_train_on_batch_multi_io(self):
        x1 = torch.rand(ModelMultiIOTest.batch_size, 1)
        x2 = torch.rand(ModelMultiIOTest.batch_size, 1)
        y1 = torch.rand(ModelMultiIOTest.batch_size, 1)
        y2 = torch.rand(ModelMultiIOTest.batch_size, 1)
        loss = self.model.train_on_batch((x1, x2), (y1, y2))
        self.assertEqual(type(loss), float)

    def test_ndarray_train_on_batch_multi_io(self):
        x1 = np.random.rand(ModelMultiIOTest.batch_size, 1).astype(np.float32)
        x2 = np.random.rand(ModelMultiIOTest.batch_size, 1).astype(np.float32)
        y1 = np.random.rand(ModelMultiIOTest.batch_size, 1).astype(np.float32)
        y2 = np.random.rand(ModelMultiIOTest.batch_size, 1).astype(np.float32)
        loss = self.model.train_on_batch((x1, x2), (y1, y2))
        self.assertEqual(type(loss), float)

    def test_evaluate_with_pred_multi_io(self):
        x = (torch.rand(ModelMultiIOTest.evaluate_dataset_len, 1), torch.rand(ModelMultiIOTest.evaluate_dataset_len, 1))
        y = (torch.rand(ModelMultiIOTest.evaluate_dataset_len, 1), torch.rand(ModelMultiIOTest.evaluate_dataset_len, 1))
        # We also test the unpacking.
        _, pred_y = self.model.evaluate(x, y, batch_size=ModelMultiIOTest.batch_size, return_pred=True)
        for pred in pred_y:
            self.assertEqual(pred.shape, (ModelMultiIOTest.evaluate_dataset_len, 1))

    def test_tensor_evaluate_on_batch_multi_io(self):
        y = (torch.rand(ModelMultiIOTest.batch_size, 1), torch.rand(ModelMultiIOTest.batch_size, 1))
        x = (torch.rand(ModelMultiIOTest.batch_size, 1), torch.rand(ModelMultiIOTest.batch_size, 1))
        loss = self.model.evaluate_on_batch(x, y)
        self.assertEqual(type(loss), float)

    def test_predict_with_np_array_multi_io(self):
        x1 = np.random.rand(ModelMultiIOTest.evaluate_dataset_len, 1).astype(np.float32)
        x2 = np.random.rand(ModelMultiIOTest.evaluate_dataset_len, 1).astype(np.float32)
        x = (x1, x2)
        pred_y = self.model.predict(x, batch_size=ModelMultiIOTest.batch_size)
        for pred in pred_y:
            self.assertEqual(pred.shape, (ModelMultiIOTest.evaluate_dataset_len, 1))

    def test_predict_generator_multi_io(self):
        num_steps = 10
        generator = some_data_tensor_generator_multi_io(ModelMultiIOTest.batch_size)
        generator = (x for x, _ in generator)
        pred_y = self.model.predict_generator(generator, steps=num_steps)

        for pred in pred_y:
            self.assertEqual(pred.shape, (num_steps * ModelMultiIOTest.batch_size, 1))

    def test_tensor_predict_on_batch_multi_io(self):
        x = (torch.rand(ModelMultiIOTest.batch_size, 1), torch.rand(ModelMultiIOTest.batch_size, 1))
        pred_y = self.model.predict_on_batch(x)
        self._test_size_and_type_for_generator(pred_y, (ModelMultiIOTest.batch_size, 1))



import inspect
import sys
from typing import Dict, BinaryIO

import torch.optim.lr_scheduler
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

from .callbacks import Callback


class _PyTorchLRSchedulerWrapper(Callback):
    """
    Default class for the LR scheduling callback. Proposes default comportment for the scheduler
    loading and saving as well as for the epoch end handling.
    """

    def __init__(self, torch_lr_scheduler, *args, **kwargs):
        super().__init__()
        if len(args) > 0 and isinstance(args[0], Optimizer):
            raise ValueError(
                "In the LR scheduler callbacks, the optimizer is "
                "automatically passed to the PyTorch's LR scheduler. "
                "You must remove it from the arguments."
            )
        self.args = args
        self.kwargs = kwargs
        self.scheduler = None
        self.state_to_load = None
        self.torch_lr_scheduler = torch_lr_scheduler

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        self.scheduler.step()

    def on_train_begin(self, logs: Dict):
        self.scheduler = self.torch_lr_scheduler(self.model.optimizer, *self.args, **self.kwargs)

        # Load state if the scheduler was not initialized when the user asked
        # to load its state
        if self.state_to_load is not None:
            self.load_state_dict(self.state_to_load)
            self.state_to_load = None

    def load_state_dict(self, state_dict):
        if self.scheduler is not None:
            self.scheduler.load_state_dict(state_dict)
        else:
            self.state_to_load = state_dict

    def state_dict(self):
        return self.scheduler.state_dict()

    def load_state(self, f: BinaryIO):
        self.load_state_dict(torch.load(f, map_location='cpu'))

    def save_state(self, f: BinaryIO):
        torch.save(self.state_dict(), f)


def new_init(torch_lr_scheduler):
    def f(self, *args, **kwargs):
        super(type(self), self).__init__(torch_lr_scheduler, *args, **kwargs)

    return f


for name, module_cls in torch.optim.lr_scheduler.__dict__.items():
    if inspect.isclass(module_cls) and issubclass(module_cls, _LRScheduler) and module_cls != _LRScheduler:
        _new_cls = type(
            name,
            (_PyTorchLRSchedulerWrapper,),
            {
                '__init__': new_init(module_cls),
                '__doc__': f"""
                            See:
                                :class:`~torch.optim.lr_scheduler.{name}`
                            """,
            },
        )
        setattr(sys.modules[__name__], name, _new_cls)


class ReduceLROnPlateau(_PyTorchLRSchedulerWrapper):
    """
    Args:
        monitor (str): The quantity to monitor. (Default value = 'val_loss')
    See:
        :class:`~torch.optim.lr_scheduler.ReduceLROnPlateau`
    """

    def __init__(self, *args, monitor: str = 'val_loss', **kwargs):
        super().__init__(torch_lr_scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau, *args, **kwargs)
        self.monitor = monitor

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        self.scheduler.step(logs[self.monitor])



import itertools
from typing import Dict, Callable

from .callbacks import Callback
from .color_formatting import ColorProgress


class ProgressionCallback(Callback):
    """
    Default progression callback used in :class:`~poutyne.Model`. You can use the ``progress_options``
    in :class:`~poutyne.Model` instead of instantiating this callback. If you choose to use this callback
    anyway, make sure to pass ``verbose=False`` to :func:`~poutyne.Model.fit()` or
    :func:`~poutyne.Model.fit_generator()`.

    Args:
        coloring (Union[bool, Dict], optional): If bool, whether to display the progress of the training with
            default colors highlighting.
            If Dict, the field and the color to use as `colorama <https://pypi.org/project/colorama/>`_ . The fields
            are ``text_color``, ``ratio_color``, ``metric_value_color``, ``time_color`` and ``progress_bar_color``.
            In both case, will be ignore if verbose is set to False.
            (Default value = True)
        progress_bar (bool): Whether or not to display a progress bar showing the epoch progress.
            Note that if the size of the output text with the progress bar is larger than the shell output size,
            the formatting could be impacted (a line for every step).
            (Default value = True)
        equal_weights (bool): Whether or not the duration of each step is weighted equally when computing the
            average time of the steps and, thus, the ETA. By default, newer step times receive more weights than
            older step times. Set this to true to have equal weighting instead.
        show_on_valid (bool): Whether or not to display the progression during the validation phase.
            (Default value = True)
        show_every_n_train_steps (Union[str, int]): Show a subset of the training steps. If ``'all'``, show all steps.
            If ``'none'``, do not show the steps (i.e. only show the stats at the end of the epoch). If an integer
            ``n``, only show every n-th steps. (Default value = 'all').
        show_every_n_valid_steps (Union[str, int]): Show a subset of the validation steps. If ``'all'``, show all steps.
            If ``'none'``, do not show the steps (i.e. only show the stats at the end of the epoch). If an integer
            ``n``, only show every n-th steps. (Default value = 'all').
        show_every_n_test_steps (Union[str, int]): Show a subset of the testing steps. If ``'all'``, show all steps.
            If ``'none'``, do not show the steps (i.e. only show the stats at the end of the testing). If an integer
            ``n``, show only every n-th steps. (Default value = 'all').
    """

    EVERY_N_STEPS_CHOICES = ['all', 'none']

    def __init__(
        self,
        *,
        coloring=True,
        progress_bar=True,
        equal_weights=False,
        show_on_valid=True,
        show_every_n_train_steps='all',
        show_every_n_valid_steps='all',
        show_every_n_test_steps='all',
    ) -> None:
        super().__init__()
        self.color_progress = ColorProgress(coloring=coloring)
        self.progress_bar = progress_bar
        self.equal_weights = equal_weights
        self.show_on_valid = show_on_valid
        self.step_times_weighted_sum = 0.0

        assert (
            isinstance(show_every_n_train_steps, int)
            or show_every_n_train_steps in ProgressionCallback.EVERY_N_STEPS_CHOICES
        )
        assert (
            isinstance(show_every_n_valid_steps, int)
            or show_every_n_valid_steps in ProgressionCallback.EVERY_N_STEPS_CHOICES
        )
        assert (
            isinstance(show_every_n_test_steps, int)
            or show_every_n_test_steps in ProgressionCallback.EVERY_N_STEPS_CHOICES
        )
        self.show_every_n_train_steps = show_every_n_train_steps
        self.show_every_n_valid_steps = show_every_n_valid_steps
        self.show_every_n_test_steps = show_every_n_test_steps

        self.train_last_step = None
        self.valid_last_step = None

    def set_params(self, params: Dict):
        super().set_params(params)
        self._train_steps = self.params['steps']
        self._valid_steps = self.params.get('valid_steps')
        self._test_steps = self.params['steps']
        self._predict_steps = self.params['steps']

    def _set_progress_bar(self):
        if self.progress_bar and self.steps is not None:
            self.color_progress.set_progress_bar(self.steps)
        elif self.progress_bar and self.steps is None:
            # Specific case where we don't have steps for the training
            # but we do during valid so we create a progress bar and
            # when we return to the train, it's all messed up
            self.color_progress.close_progress_bar()

    def on_train_begin(self, logs: Dict) -> None:
        self.metrics = ['loss'] + self.model.metrics_names
        self.epochs = self.params['epochs']
        self.steps = self._train_steps

    def on_valid_begin(self, logs: Dict) -> None:
        if self.show_on_valid:
            self.step_times_weighted_sum = 0.0

            self.metrics = ['loss'] + self.model.metrics_names
            self.steps = self._valid_steps

            self._set_progress_bar()

            self.color_progress.on_valid_begin()

    def on_test_begin(self, logs: Dict) -> None:
        self.step_times_weighted_sum = 0.0

        self.metrics = ['loss'] + self.model.metrics_names
        self.steps = self._test_steps

        self._set_progress_bar()

        self.color_progress.on_test_begin()

    def on_predict_begin(self, logs: Dict) -> None:
        self.step_times_weighted_sum = 0.0

        self.metrics = []
        self.steps = self._predict_steps

        self._set_progress_bar()

        self.color_progress.on_predict_begin()

    def on_epoch_begin(self, epoch_number: int, logs: Dict) -> None:
        self.step_times_weighted_sum = 0.0
        self.epoch_number = epoch_number
        self.steps = self._train_steps

        self._set_progress_bar()

        self.color_progress.on_epoch_begin(epoch_number=self.epoch_number, epochs=self.epochs)

    def on_epoch_end(self, epoch_number: int, logs: Dict) -> None:
        self.steps = self._train_steps
        epoch_total_time = logs['time']

        metrics_str = self._get_metrics_string(logs)
        self.color_progress.on_epoch_end(
            total_time=epoch_total_time,
            train_last_steps=self.train_last_step,
            valid_last_steps=self.valid_last_step,
            metrics_str=metrics_str,
        )

    def on_test_end(self, logs: Dict) -> None:
        test_total_time = logs['time']
        progress_fun = self.color_progress.on_test_end

        self._end_progress(logs, test_total_time, progress_fun)

    def on_predict_end(self, logs: Dict) -> None:
        predict_total_time = logs['time']
        progress_fun = self.color_progress.on_predict_end

        self._end_progress(logs, predict_total_time, progress_fun)

    def _test_show_batch_end(self, show_every_n_steps_flag, batch_number):
        return show_every_n_steps_flag == 'all' or (
            isinstance(show_every_n_steps_flag, int) and batch_number % show_every_n_steps_flag == 0
        )

    def on_train_batch_end(self, batch_number: int, logs: Dict) -> None:
        train_step_times_rate = self._compute_step_times_rate(batch_number, logs)
        progress_batch_end_fun = self.color_progress.on_train_batch_end

        do_print = self._test_show_batch_end(self.show_every_n_train_steps, batch_number)
        self._batch_end_progress(
            logs=logs,
            step_times_rate=train_step_times_rate,
            batch_number=batch_number,
            func=progress_batch_end_fun,
            do_print=do_print,
        )
        self.train_last_step = batch_number

    def on_valid_batch_end(self, batch_number: int, logs: Dict) -> None:
        if self.show_on_valid:
            valid_step_times_rate = self._compute_step_times_rate(batch_number, logs)
            progress_batch_end_fun = self.color_progress.on_valid_batch_end

            do_print = self._test_show_batch_end(self.show_every_n_valid_steps, batch_number)
            self._batch_end_progress(
                logs=logs,
                step_times_rate=valid_step_times_rate,
                batch_number=batch_number,
                func=progress_batch_end_fun,
                do_print=do_print,
            )

        self.valid_last_step = batch_number

    def on_test_batch_end(self, batch_number: int, logs: Dict) -> None:
        test_step_times_rate = self._compute_step_times_rate(batch_number, logs)
        progress_batch_end_fun = self.color_progress.on_test_batch_end

        do_print = self._test_show_batch_end(self.show_every_n_test_steps, batch_number)
        self._batch_end_progress(
            logs=logs,
            step_times_rate=test_step_times_rate,
            batch_number=batch_number,
            func=progress_batch_end_fun,
            do_print=do_print,
        )

    def on_predict_batch_end(self, batch_number: int, logs: Dict) -> None:
        predict_step_times_rate = self._compute_step_times_rate(batch_number, logs)
        progress_batch_end_fun = self.color_progress.on_predict_batch_end

        do_print = self._test_show_batch_end(self.show_every_n_test_steps, batch_number)
        self._batch_end_progress(
            logs=logs,
            step_times_rate=predict_step_times_rate,
            batch_number=batch_number,
            func=progress_batch_end_fun,
            do_print=do_print,
        )

    def _get_metrics_string(self, logs: Dict) -> str:
        train_metrics_str_gen = (f'{k}: {logs[k]:f}' for k in self.metrics if logs.get(k) is not None)
        val_metrics_str_gen = (
            f"{'val_' + k}: {logs['val_' + k]:f}" for k in self.metrics if logs.get('val_' + k) is not None
        )
        test_metrics_str_gen = (
            f"{'test_' + k}: {logs['test_' + k]:f}" for k in self.metrics if logs.get('test_' + k) is not None
        )
        return ', '.join(itertools.chain(train_metrics_str_gen, val_metrics_str_gen, test_metrics_str_gen))

    def _compute_step_times_rate(self, batch_number: int, logs: Dict) -> float:
        if self.equal_weights:
            self.step_times_weighted_sum += logs['time']
            step_times_rate = self.step_times_weighted_sum / batch_number
        else:
            self.step_times_weighted_sum += batch_number * logs['time']
            normalizing_factor = (batch_number * (batch_number + 1)) / 2
            step_times_rate = self.step_times_weighted_sum / normalizing_factor
        return step_times_rate

    def _end_progress(self, logs: Dict, total_time: float, func: Callable) -> None:
        """
        Update the progress at the end of a test or valid phase.
        """
        metrics_str = self._get_metrics_string(logs)
        if self.steps is not None:
            func(total_time=total_time, steps=self.steps, metrics_str=metrics_str)
        else:
            func(total_time=total_time, steps=self.last_step, metrics_str=metrics_str)

    def _batch_end_progress(
        self, *, logs: Dict, step_times_rate: float, batch_number: int, func: Callable, do_print: bool
    ) -> None:
        """
        Update the progress at the end of train, valid or test batch.
        """
        metrics_str = self._get_metrics_string(logs)
        if self.steps is not None:
            remaining_time = step_times_rate * (self.steps - batch_number)
            func(
                remaining_time=remaining_time,
                batch_number=batch_number,
                metrics_str=metrics_str,
                steps=self.steps,
                do_print=do_print,
            )
        else:
            func(remaining_time=step_times_rate, batch_number=batch_number, metrics_str=metrics_str, do_print=do_print)
            self.last_step = batch_number


class EpochProgressionCallback(Callback):
    EVERY_N_EPOCHS_CHOICES = ['all', 'none']

    def __init__(self, *, coloring=True, show_every_n_epochs='all') -> None:
        super().__init__()
        self.color_progress = ColorProgress(coloring=coloring)

        assert (
            isinstance(show_every_n_epochs, int)
            or show_every_n_epochs in EpochProgressionCallback.EVERY_N_EPOCHS_CHOICES
        )
        self.show_every_n_epochs = show_every_n_epochs

    def set_params(self, params: Dict):
        super().set_params(params)
        self._train_steps = self.params['steps']
        self._valid_steps = self.params.get('valid_steps')

    def on_train_begin(self, logs: Dict) -> None:
        self.metrics = ['loss'] + self.model.metrics_names
        self.epochs = self.params['epochs']

    def _test_show_epoch(self, epoch_number):
        return self.show_every_n_epochs != 'none' and (
            self.show_every_n_epochs == 'all' or epoch_number % self.show_every_n_epochs == 0
        )

    def on_epoch_begin(self, epoch_number: int, logs: Dict) -> None:
        if self._test_show_epoch(epoch_number):
            self.color_progress.on_epoch_begin(epoch_number=epoch_number, epochs=self.epochs)

    def on_epoch_end(self, epoch_number: int, logs: Dict) -> None:
        if self._test_show_epoch(epoch_number):
            self.color_progress.on_epoch_end(
                total_time=logs['time'],
                train_last_steps=self._train_steps,
                valid_last_steps=self._valid_steps,
                metrics_str=self._get_metrics_string(logs),
            )

    def _get_metrics_string(self, logs: Dict) -> str:
        train_metrics_str_gen = (f'{k}: {logs[k]:f}' for k in self.metrics if logs.get(k) is not None)
        val_metrics_str_gen = (
            f"{'val_' + k}: {logs['val_' + k]:f}" for k in self.metrics if logs.get('val_' + k) is not None
        )
        return ', '.join(itertools.chain(train_metrics_str_gen, val_metrics_str_gen))



from typing import Dict, Optional

from .callbacks import Callback, CallbackList


class DelayCallback(Callback):
    """
    Delays one or many callbacks for a certain number of epochs or number of batches. If both
    ``epoch_delay`` and ``batch_delay`` are provided, the biggest has precedence.

    Args:
        callbacks (Callback, List[Callback]): A callback or a list of callbacks to delay.
        epoch_delay (int, optional): Number of epochs to delay.
        batch_delay (int, optional): Number of batches to delay. The number of batches can span many
            epochs. When the batch delay expires (i.e. there are more than `batch_delay` done), the
            :func:`~poutyne.Callback.on_epoch_begin()` method is called on
            the callback(s) before the :func:`~poutyne.Callback.on_train_batch_begin()` method.
    """

    def __init__(self, callbacks: Callback, *, epoch_delay: Optional[int] = None, batch_delay: Optional[int] = None):
        super().__init__()
        if isinstance(callbacks, CallbackList):
            self.callbacks = callbacks
        elif isinstance(callbacks, list):
            self.callbacks = CallbackList(callbacks)
        else:
            self.callbacks = CallbackList([callbacks])

        self.epoch_delay = epoch_delay if epoch_delay else 0
        self.batch_delay = batch_delay if batch_delay else 0

    def set_params(self, params: Dict):
        self.callbacks.set_params(params)

    def set_model(self, model):
        self.callbacks.set_model(model)

    def on_epoch_begin(self, epoch_number: int, logs: Dict):
        self.current_epoch = epoch_number
        if self.has_delay_passed():
            self.has_on_epoch_begin_been_called = True
            self.callbacks.on_epoch_begin(epoch_number, logs)

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        if self.has_delay_passed():
            self.callbacks.on_epoch_end(epoch_number, logs)

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        self.batch_counter += 1
        if self.has_delay_passed():
            if not self.has_on_epoch_begin_been_called:
                self.has_on_epoch_begin_been_called = True
                self.callbacks.on_epoch_begin(self.current_epoch, logs)
            self.callbacks.on_train_batch_begin(batch_number, logs)

    def on_train_batch_end(self, batch_number: int, logs: Dict):
        if self.has_delay_passed():
            self.callbacks.on_train_batch_end(batch_number, logs)

    def on_backward_end(self, batch_number: int):
        if self.has_delay_passed():
            self.callbacks.on_backward_end(batch_number)

    def on_train_begin(self, logs: Dict):
        self.current_epoch = 0
        self.batch_counter = 0
        self.has_on_epoch_begin_been_called = False
        self.callbacks.on_train_begin(logs)

    def on_train_end(self, logs: Dict):
        self.callbacks.on_train_end(logs)

    def has_delay_passed(self):
        return self.current_epoch > self.epoch_delay and self.batch_counter > self.batch_delay



"""
The source code of this file was copied from the Keras project, and has been modified. All modifications
made from the original source code are under the LGPLv3 license.

COPYRIGHT

All contributions by François Chollet:
Copyright (c) 2015, François Chollet.
All rights reserved.

All contributions by Google:
Copyright (c) 2015, Google, Inc.
All rights reserved.

All contributions by Microsoft:
Copyright (c) 2017, Microsoft, Inc.
All rights reserved.

All other contributions:
Copyright (c) 2015 - 2017, the respective contributors.
All rights reserved.

Each contributor holds copyright over their respective contributions. The project versioning (Git)
records all such contribution source information.

LICENSE

The MIT License (MIT)

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES
OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import warnings
from typing import IO, Dict, Optional, Callable

from ._utils import atomic_lambda_save
from .callbacks import Callback


class PeriodicSaveCallback(Callback):
    """
    Write a file (or checkpoint) after every epoch. `filename` can contain named formatting options,
    which will be filled the value of `epoch` and keys in `logs` (passed in `on_epoch_end`). For
    example: if `filename` is `weights.{epoch:02d}-{val_loss:.2f}.txt`, then `save_file()` will be
    called with a file descriptor for a file with the epoch number and the validation loss in the
    filename.

    By default, the file is written atomically to the specified filename so that the training can
    be killed and restarted later using the same filename for periodic file saving. To do so, a
    temporary file is created with the name of `filename + '.tmp'` and is then moved to the final
    destination after the checkpoint is done. The ``temporary_filename`` argument allows to change the
    path of this temporary file.

    Args:
        filename (str): Path to save the model file.
        monitor (str): Quantity to monitor.
            (Default value = 'val_loss')
        verbose (bool): Whether to display a message when saving and restoring a checkpoint.
            (Default value = False)
        save_best_only (bool): If `save_best_only` is true, the latest best model according to the
            quantity monitored will not be overwritten.
            (Default value = False)
        keep_only_last_best (bool): Whether only the last saved best checkpoint is kept. Applies only when
             `save_best_only` is true.
             (Default value = False)
        restore_best (bool): If `restore_best` is true, the model will be reset to the last best checkpoint done.
            This option only works when `save_best_only` is also true. (Default value = False)
        mode (str): One of {'min', 'max'}.
            If `save_best_only` is true, the decision to overwrite the current save file is made based
            on either the maximization or the minimization of the monitored quantity. For
            `val_accuracy`, this should be `max`, for `val_loss` this should be `min`, etc.
            (Default value = 'min')
        period (int): Interval (number of epochs) between checkpoints.
            (Default value = 1)
        temporary_filename (str, optional): Temporary filename for the checkpoint so that the
            last checkpoint can be written atomically. See the ``atomic_write`` argument.
        atomic_write (bool): Whether to write atomically the checkpoint. See the description above
            for details.
            (Default value = True)
        open_mode (str): ``mode`` option passed to :func:`open()`.
            (Default value = 'wb')
    """

    def __init__(
        self,
        filename: str,
        *,
        monitor: str = 'val_loss',
        mode: str = 'min',
        save_best_only: bool = False,
        keep_only_last_best: bool = False,
        restore_best: bool = False,
        period: int = 1,
        verbose: bool = False,
        temporary_filename: Optional[str] = None,
        atomic_write: bool = True,
        open_mode: str = 'wb',
        read_mode: str = 'rb',
    ):
        super().__init__()
        self.filename = filename
        self.monitor = monitor
        self.verbose = verbose
        self.save_best_only = save_best_only
        self.keep_only_last_best = keep_only_last_best
        self.restore_best = restore_best
        self.temporary_filename = temporary_filename
        self.atomic_write = atomic_write
        self.open_mode = open_mode
        self.read_mode = read_mode
        self.best_filename = None

        if self.keep_only_last_best and not self.save_best_only:
            raise ValueError("The 'keep_only_last_best' argument only works when 'save_best_only' is also true.")
        if self.restore_best and not self.save_best_only:
            raise ValueError("The 'restore_best' argument only works when 'save_best_only' is also true.")

        if self.save_best_only:
            if mode not in ['min', 'max']:
                raise ValueError(f"Invalid mode '{mode}'")
            if mode == 'min':
                self.monitor_op = lambda x, y: x < y
                self.current_best = float('Inf')
            elif mode == 'max':
                self.monitor_op = lambda x, y: x > y
                self.current_best = -float('Inf')

        self.period = period

    def save_file(self, fd: IO, epoch_number: int, logs: Dict):
        raise NotImplementedError

    def _save_file(self, filename: str, epoch_number: int, logs: Dict):
        atomic_lambda_save(
            filename,
            self.save_file,
            (epoch_number, logs),
            temporary_filename=self.temporary_filename,
            open_mode=self.open_mode,
            atomic=self.atomic_write,
        )

    def on_epoch_end(self, epoch_number: int, logs: Dict):
        filename = self.filename.format_map(logs)

        if self.save_best_only:
            if self.monitor_op(logs[self.monitor], self.current_best):
                old_best = self.current_best
                self.current_best = logs[self.monitor]
                old_best_filename = self.best_filename
                self.best_filename = filename

                if self.verbose:
                    print(
                        f'Epoch {epoch_number:d}: {self.monitor} improved from {old_best:0.5f} '
                        f'to {self.current_best:0.5f}, saving file to {self.best_filename}'
                    )
                self._save_file(self.best_filename, epoch_number, logs)
                if (
                    self.keep_only_last_best
                    and self.best_filename != old_best_filename
                    and old_best_filename is not None
                ):
                    os.remove(old_best_filename)
        elif epoch_number % self.period == 0:
            if self.verbose:
                print(f'Epoch {epoch_number:d}: saving file to {filename}')
            self._save_file(filename, epoch_number, logs)

    def restore(self, fd: IO):
        raise NotImplementedError

    def on_train_end(self, logs: Dict):
        if self.restore_best:
            if self.best_filename is not None:
                if self.verbose:
                    print(f'Restoring data from {self.best_filename}')
                # pylint: disable=unspecified-encoding
                open_kwargs = dict(encoding='utf-8') if 'b' not in self.read_mode else {}
                with open(self.best_filename, self.read_mode, **open_kwargs) as fd:
                    self.restore(fd)
            else:
                warnings.warn('No data to restore!')


class PeriodicSaveLambda(PeriodicSaveCallback):
    """
    Call a lambda with a file descriptor after every epoch. See
    :class:`~poutyne.PeriodicSaveCallback` for the arguments' descriptions.

    Args:
        func (Callable[[fd, int, dict], None]): The lambda that will be called with a file descriptor, the
            epoch number and the epoch logs.
        restore (Callable[[fd], None]): The lambda that will be called with a file descriptor to restore
            the state if necessary.

    See:
        :class:`~poutyne.PeriodicSaveCallback`
    """

    def __init__(self, func: Callable, *args, restore: Optional[Callable] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.func = func
        self._restore = restore

    def save_file(self, fd: IO, epoch_number: int, logs: Dict):
        self.func(fd, epoch_number, logs)

    def restore(self, fd: IO):
        if self._restore is not None:
            self._restore(fd)



import unittest
from unittest import TestCase

import torch
import torch.nn as nn

from poutyne import torch_to_numpy, Model, BestModelRestore
from tests.framework.tools import some_data_generator


class BestModelRestoreTest(TestCase):
    batch_size = 20

    def setUp(self):
        torch.manual_seed(42)
        self.pytorch_network = nn.Linear(1, 1)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=1e-3)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

    def test_integration(self):
        train_gen = some_data_generator(20)
        valid_gen = some_data_generator(20)
        model_restore = BestModelRestore(monitor='val_loss', verbose=True)
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[model_restore])

    def test_basic_restore(self):
        model_restore = BestModelRestore(monitor='val_loss')

        val_losses = [3, 2, 8, 5, 4]
        best_epoch = 2
        self._test_restore_with_val_losses(model_restore, val_losses, best_epoch)

    def test_save_best_only_with_max(self):
        model_restore = BestModelRestore(monitor='val_loss', mode='max')

        val_losses = [3, 2, 8, 5, 4]
        best_epoch = 3
        self._test_restore_with_val_losses(model_restore, val_losses, best_epoch)

    def _test_restore_with_val_losses(self, checkpointer, val_losses, best_epoch):
        generator = some_data_generator(BestModelRestoreTest.batch_size)

        best_epoch_weights = None
        checkpointer.set_params({'epochs': len(val_losses), 'steps': 1})
        checkpointer.set_model(self.model)
        checkpointer.on_train_begin({})
        for epoch, val_loss in enumerate(val_losses, 1):
            checkpointer.on_epoch_begin(epoch, {})
            checkpointer.on_train_batch_begin(1, {})
            loss = self._update_model(generator)
            checkpointer.on_train_batch_end(1, {'batch': 1, 'size': BestModelRestoreTest.batch_size, 'loss': loss})
            checkpointer.on_epoch_end(epoch, {'epoch': epoch, 'loss': loss, 'val_loss': val_loss})
            if epoch == best_epoch:
                best_epoch_weights = torch_to_numpy(self.model.get_weight_copies())
        checkpointer.on_train_end({})

        final_weights = torch_to_numpy(self.model.get_weight_copies())
        self.assertEqual(best_epoch_weights, final_weights)

    def _update_model(self, generator):
        self.pytorch_network.zero_grad()

        x, y = next(generator)
        pred_y = self.pytorch_network(x)
        loss = self.loss_function(pred_y, y)
        loss.backward()

        self.optimizer.step()

        return float(loss)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod

import torch.nn as nn


class EpochMetric(ABC, nn.Module):
    """
    The abstract class representing a epoch metric which can be accumulated at each batch and calculated at the end
    of the epoch.
    """

    @abstractmethod
    def forward(self, y_pred, y_true) -> None:
        """
        To define the behavior of the metric when called.

        Args:
            y_pred: The prediction of the model.
            y_true: Target to evaluate the model.
        """
        pass

    @abstractmethod
    def get_metric(self):
        """
        Compute and return the metric. Should not modify the state of
        the epoch metric.
        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        The information kept for the computation of the metric is cleaned so
        that a new epoch can be done.
        """
        pass



import os
import itertools
from typing import Any, Dict, List, Optional, Tuple, Union

try:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MaxNLocator

    matplotlib = True
except ImportError:
    matplotlib = False

try:
    import pandas as pd
except ImportError:
    pd = None

from poutyne import is_in_jupyter_notebook

jupyter = is_in_jupyter_notebook()


def _raise_error_if_matplotlib_not_there():
    if not matplotlib:
        raise ImportError("matplotlib needs to be installed to use this function.")


def _none_to_iterator(value, repeat=None):
    return value if value is not None else itertools.repeat(repeat)


def _assert_list_length_with_num_metrics(l, metrics, name):
    if l is not None and len(l) != len(metrics):
        raise ValueError(
            f"A {name} was not provided for each metric. " f"Got {len(l)} {name}s for {len(metrics)} metrics."
        )


def _infer_metrics(history, metrics):
    if metrics is None:
        if pd is not None and isinstance(history, pd.DataFrame):
            cols = list(history.columns)
        else:
            cols = list(history[0].keys())
        metrics = [col for col in cols if col != 'epoch' and not col.startswith('val_')]
    return metrics


def _get_figs_and_axes(axes, num_axes, fig_kwargs):
    figs = ()
    if axes is None:
        fig_kwargs = fig_kwargs if fig_kwargs is not None else {}
        figs, axes = zip(*(plt.subplots(**fig_kwargs) for _ in range(num_axes)))
    return figs, axes


def _save_figs(figs, metrics, *, filename_template, directory, extensions):
    save_template = filename_template
    if directory is not None:
        os.makedirs(directory, exist_ok=True)
        save_template = os.path.join(directory, filename_template)

    for fig, metric in zip(figs, metrics):
        for ext in extensions:
            filename = save_template.format(metric=metric) + f'.{ext}'
            fig.savefig(filename)


def _show_figs(figs):
    for fig in figs:
        fig.show()


def _close_figs(figs):
    for fig in figs:
        plt.close(fig)


def plot_history(
    history: Union[List[Dict[str, Union[float, int]]], 'pd.DataFrame'],
    *,
    metrics: Optional[List[str]] = None,
    labels: Optional[List[str]] = None,
    titles: Optional[Union[List[str], str]] = None,
    axes: Optional[List['matplotlib.axes.Axes']] = None,
    show: bool = True,
    save: bool = False,
    save_filename_template: str = '{metric}',
    save_directory: Optional[str] = None,
    save_extensions: Union[List[str], Tuple[str]] = ('png',),
    close: Optional[bool] = None,
    fig_kwargs: Optional[Dict[str, Any]] = None,
):
    """
    Plot the training history in matplotlib. By default, all metrics are plotted.

    Args:
        history (Union[List[Dict[str, Union[float, int]]], pandas.DataFrame]): The training history to plot. Can be
            either a list of dictionary as returned by :func:`~poutyne.Model.fit()` or a Pandas DataFrame as read from a
            CSV output by the :class:`~poutyne.CSVLogger` callback.
        metrics (Optional[List[str]], optional): The list of metrics for which to output the plot. By default, every
            metric in the history is used.
        labels (Optional[List[str]], optional): A list of labels to use for each metric. Must be of the same length as
            ``metrics``. By default, the names in the history are used.
        titles (Optional[Union[List[str], str]], optional): A title or a list of titles to use for each metric. If a
            list, must be of the same length as ``metrics``. If a string, the same title will be used for all plots. By
            default, there is no title.
        axes (Optional[List[matplotlib.axes.Axes]], optional): A list of matplotlib :class:`~matplotlib.axes.Axes` to
            use for each metric. Must be of the same length as ``metrics``. By default, a new figure and an new axe is
            created for each plot.
        show (bool, optional): Whether to show the plots. Defaults to True.
        save (bool, optional): Whether to save the plots. Defaults to False.
        save_filename_template (str, optional): The filename without extension for saving the plot. Should contain
            ``{metric}`` somewhere in it or all the plots will overwrite each other. Defaults to ``'{metric}'``.
        save_directory (Optional[str], optional): The directory to save the plots. Default to the current directory.
        save_extensions (Union[List[str], Tuple[str]], optional): A list of extensions under which to save the plots.
            Defaults to `('png', )`.
        close (Optional[bool], optional): Whether to close the matplotlib figures. By default, the figures are closed
            except when in Jupyter notebooks.
        fig_kwargs (Optional[Dict[str, Any]], optional): Any keyword arguments to pass to
            :func:`~matplotlib.pyplot.subplots`.

    Returns:
        Tuple[List[matplotlib.figure.Figure], List[matplotlib.axes.Axes]]: A tuple ``(figs, axes)``  where ``figs`` is
        the list of instanciated matplotlib :class:`~matplotlib.figure.Figure` and ``axes`` is a list of instanciated
        matplotlib :class:`~matplotlib.figure.Axes`.
    """

    # pylint: disable=too-many-locals
    _raise_error_if_matplotlib_not_there()

    metrics = _infer_metrics(history, metrics)

    _assert_list_length_with_num_metrics(labels, metrics, 'label')

    if isinstance(titles, str):
        titles = [titles] * len(metrics)
    else:
        _assert_list_length_with_num_metrics(titles, metrics, 'title')

    _assert_list_length_with_num_metrics(axes, metrics, 'axe')

    labels = _none_to_iterator(labels)
    titles = _none_to_iterator(titles, repeat='')

    figs, axes = _get_figs_and_axes(axes, len(metrics), fig_kwargs)
    for metric, label, title, ax in zip(metrics, labels, titles, axes):
        plot_metric(history, metric, label=label, title=title, ax=ax)

    if save:
        _save_figs(
            figs,
            metrics,
            filename_template=save_filename_template,
            directory=save_directory,
            extensions=save_extensions,
        )

    if show:
        _show_figs(figs)

    if close is None:
        close = not jupyter

    if close:
        _close_figs(figs)

    return figs, axes


def plot_metric(
    history: Union[List[Dict[str, Union[float, int]]], 'pd.DataFrame'],
    metric: str,
    *,
    label: Optional[str] = None,
    title: str = '',
    ax: Optional['matplotlib.axes.Axes'] = None,
):
    """
    Plot the training history in matplotlib for a given metric.

    Args:
        history (Union[List[Dict[str, Union[float, int]]], pd.DataFrame]): The training history to plot. Can be
            either a list of dictionary as returned by :func:`~poutyne.Model.fit()` or a Pandas DataFrame as read from a
            CSV output by the :class:`~poutyne.CSVLogger` callback.
        metric (str): The metric for which to output the plot.
        label (str, Optional[str]): A label for the metric. By default, the label is the same as the name of the metric.
        title (str, optional): A title for the plot. By default, no title.
        ax (Optional[matplotlib.axes.Axes], optional): A matplotlib :class:`~matplotlib.axes.Axes` to use. By default,
            the current axe is used.
    """
    _raise_error_if_matplotlib_not_there()

    if ax is None:
        ax = plt.gca()

    if label is None:
        train_label = metric
        valid_label = 'val_' + metric
    else:
        train_label = 'Training ' + label
        valid_label = 'Validation ' + label

    val_metric_values = None
    if pd is not None and isinstance(history, pd.DataFrame):
        epochs = history['epoch']
        metric_values = history[metric]
        if 'val_' + metric in history:
            val_metric_values = history['val_' + metric]
    else:
        epochs = [entry['epoch'] for entry in history]
        metric_values = [entry[metric] for entry in history]
        if 'val_' + metric in history[0]:
            val_metric_values = [entry['val_' + metric] for entry in history]

    ax.plot(epochs, metric_values, label=train_label)
    if val_metric_values is not None:
        ax.plot(epochs, val_metric_values, label=valid_label)

    ax.set_xlabel('Epochs')
    ax.set_ylabel(label)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    ax.legend()
    ax.set_title(title)
